@@ -100,6 +100,10 @@ form:\n\
 Notifications emitted by applications can also be handled by using the \n\
 :py:func:`watch` and :py:func:`set_callback` methods.\n\
 \n\
+One may also examine and perform the available actions for the element with the \n\
+:py:func:`actions`, :py:func:`action_description`, and :py:func:`perform_action` \n\
+methods.\n\
+\n\
 Some care has been taken to manage the memory used by the CFTypeRef system, but \n\
 this has not been rigorously tested.\n\
 \n\
@@ -252,6 +256,23 @@ For example, to watch for when windows are moved or resized in an application:\n
 
 static PyObject * AccessibleElement_watch(AccessibleElement *, PyObject *);
 
+PyDoc_STRVAR(actions_docstring, "actions()\n\n\
+Retrieves the list of actions available for the element. This is similar to the \n\
+list of attributes that is returned by the :py:func:`keys` method.");
+
+static PyObject * AccessibleElement_actions(AccessibleElement *, PyObject *);
+
+PyDoc_STRVAR(action_desciption_docstring, "action_description(action_name)\n\n\
+Retrieves the description of an action available for the element. Note that \n\
+descriptions are implemented by the vendors, and so are often unhelpful.");
+
+static PyObject * AccessibleElement_action_description(AccessibleElement *, PyObject *);
+
+PyDoc_STRVAR(perform_action_docstring, "perform_action(action_name)\n\n\
+Performs the specified action on the element, if it is possible to do so.");
+
+static PyObject * AccessibleElement_perform_action(AccessibleElement *, PyObject *);
+
 /* Module functions
 ======== */
 
@@ -315,8 +336,9 @@ static void NotifcationCallback(AXObserverRef, AXUIElementRef, CFStringRef, void
 ======== */
 
 static void AccessibleElement_dealloc(AccessibleElement * self) {
-    // Use CFRelease to release for the AXUIElementRef
+    // Use CFRelease to release for the AXUIElementRef, AXObserverRef
     if (self->_ref != NULL) CFRelease(self->_ref);
+    if (self->_obs != NULL) CFRelease(self->_obs);
     Py_XDECREF(self->pid);
 #if PY_MAJOR_VERSION >= 3
     Py_TYPE(self)->tp_free((PyObject *) self);
@@ -732,16 +754,90 @@ static PyObject * AccessibleElement_watch(AccessibleElement * self, PyObject * a
     Py_RETURN_NONE;
 }
 
+static PyObject * AccessibleElement_actions(AccessibleElement * self, PyObject * args) {
+    PyObject * result = NULL;
+    CFArrayRef names;
+    AXError error = AXUIElementCopyActionNames(self->_ref, &names);
+    if (error == kAXErrorSuccess) {
+        result = parseCFTypeRef(names);
+    } else {
+        handleAXErrors("action names", error);
+    }
+
+    if (names != NULL) CFRelease(names);
+    return result;
+}
+
+static PyObject * AccessibleElement_action_description(AccessibleElement * self, PyObject * args) {
+    Py_ssize_t attribute_count = PyTuple_Size(args);
+    if (attribute_count != 1) {
+        PyErr_SetString(PyExc_ValueError, "An action name must be specified to retrieve its description.");
+        return NULL;
+    }
+    // The first argument should be a string
+    PyObject * name = PyTuple_GetItem(args, (Py_ssize_t) 0);
+    if (!name) {
+        return NULL; // PyTuple_GetItem will set an Index error.
+    }
+    char * name_string = NULL;
+    CFStringRef name_strref = CFStringFromPyString(name, &name_string);
+    if (!name_strref) return NULL; // CFStringFromPyString will set an error.
+
+    // Check for the action's description
+    CFStringRef descr;
+    AXError error = AXUIElementCopyActionDescription(self->_ref, name_strref, &descr);
+    if (name_strref != NULL) CFRelease(name_strref);
+
+    if (error != kAXErrorSuccess) {
+        handleAXErrors(name_string, error);
+        return NULL;
+    }
+    return parseCFTypeRef(descr);
+}
+
+static PyObject * AccessibleElement_perform_action(AccessibleElement * self, PyObject * args) {
+    Py_ssize_t attribute_count = PyTuple_Size(args);
+    if (attribute_count != 1) {
+        PyErr_SetString(PyExc_ValueError, "An action name must be specified.");
+        return NULL;
+    }
+    // The first argument should be a string
+    PyObject * name = PyTuple_GetItem(args, (Py_ssize_t) 0);
+    if (!name) {
+        return NULL; // PyTuple_GetItem will set an Index error.
+    }
+    char * name_string = NULL;
+    CFStringRef name_strref = CFStringFromPyString(name, &name_string);
+    if (!name_strref) return NULL; // CFStringFromPyString will set an error.
+
+    // Check for the action's description
+    AXError error = AXUIElementPerformAction(self->_ref, name_strref);
+    if (name_strref != NULL) CFRelease(name_strref);
+
+    if (error != kAXErrorSuccess) {
+        handleAXErrors(name_string, error);
+        return NULL;
+    }
+    Py_RETURN_NONE;
+}
+
 static PyMethodDef AccessibleElement_methods[] = {
+    // Attributes
     {"keys", (PyCFunction) AccessibleElement_keys, METH_NOARGS, keys_docstring},
     {"count", (PyCFunction) AccessibleElement_count, METH_VARARGS, count_docstring},
     {"get", (PyCFunction) AccessibleElement_get, METH_VARARGS, get_docstring},
     {"set", (PyCFunction) AccessibleElement_set, METH_VARARGS, set_docstring},
-    {"set_callback", (PyCFunction) AccessibleElement_set_callback, METH_VARARGS, set_callback_docstring},
-    {"set_timeout", (PyCFunction) AccessibleElement_set_timeout, METH_VARARGS, set_timeout_docstring},
     {"can_set", (PyCFunction) AccessibleElement_can_set, METH_VARARGS, can_set_docstring},
-    {"is_alive", (PyCFunction) AccessibleElement_is_alive, METH_NOARGS, is_alive_docstring},
+    // Notification API
     {"watch", (PyCFunction) AccessibleElement_watch, METH_VARARGS, watch_docstring},
+    {"set_callback", (PyCFunction) AccessibleElement_set_callback, METH_VARARGS, set_callback_docstring},
+    // Action API
+    {"actions", (PyCFunction) AccessibleElement_actions, METH_NOARGS, actions_docstring},
+    {"action_description", (PyCFunction) AccessibleElement_action_description, METH_VARARGS, action_desciption_docstring},
+    {"perform_action", (PyCFunction) AccessibleElement_perform_action, METH_VARARGS, perform_action_docstring},
+    // Misc
+    {"set_timeout", (PyCFunction) AccessibleElement_set_timeout, METH_VARARGS, set_timeout_docstring},
+    {"is_alive", (PyCFunction) AccessibleElement_is_alive, METH_NOARGS, is_alive_docstring},
     {NULL, NULL, 0, NULL}
 };
 
@@ -835,10 +931,11 @@ static PyObject * is_trusted(PyObject * self) {
 }
 
 static AccessibleElement * create_application_ref(PyObject * self, PyObject * args, PyObject * kwargs) {
-    static char *kwlist [] = {"pid", NULL};
+    static char *kwlist [] = {"pid", "force", NULL};
     pid_t pid;
+    int force = 0;
  
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i", kwlist, &pid))
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|i", kwlist, &pid, &force))
         return NULL;
 
     AXUIElementRef ref = AXUIElementCreateApplication(pid);
@@ -852,10 +949,14 @@ static AccessibleElement * create_application_ref(PyObject * self, PyObject * ar
     if (value != NULL) CFRelease(value);
     
     if (error == kAXErrorAPIDisabled) {
-        PyErr_SetString(APIDisabledError, "This PID does not respond to Accessibility requests -- perhaps Accessibility is not enabled on the system?");
+        PyErr_SetString(APIDisabledError, "The element created with this PID does not respond to Accessibility requests -- perhaps Accessibility is not enabled on the system?");
         return NULL;
-    } else if (error != kAXErrorSuccess) {
-        PyErr_SetString(PyExc_ValueError, "This PID does not seem to be associated with a valid process.");
+    } else if (error != kAXErrorSuccess && force == 0) {
+        PyErr_SetString(PyExc_ValueError, formattedMessage(
+           "The element does not respond to a request for its AXRole, which is \n\
+supposedly required of all Accessibility API-enabled objects. For this reason, \n\
+it is inadvisable to try and create an AccessibleElement for this PID. You may \n\
+override this failure case by passing ``force = True`` to this function.", error));
         return NULL;
     }
     
@@ -901,7 +1002,7 @@ static PyMethodDef methods[] = {
     {"is_enabled", (PyCFunction) is_enabled, METH_NOARGS, "is_enabled()\n\nCheck if accessibility has been enabled on the system."},
 #endif
     {"is_trusted", (PyCFunction) is_trusted, METH_NOARGS, "is_trusted()\n\nCheck if this application is a trusted process."},
-    {"create_application_ref", (PyCFunction) create_application_ref, METH_VARARGS|METH_KEYWORDS, "create_application_ref(pid)\n\nCreate an accessibile application with the given PID."},
+    {"create_application_ref", (PyCFunction) create_application_ref, METH_VARARGS|METH_KEYWORDS, "create_application_ref(pid, force = False)\n\nCreate an accessibile application with the given PID."},
     {"create_systemwide_ref", (PyCFunction) create_systemwide_ref, METH_NOARGS, "create_systemwide_ref()\n\nGet a system-wide accessible element reference."},
     {"element_at_position", (PyCFunction) element_at_position, METH_VARARGS|METH_KEYWORDS, element_at_position_docstring},
     {NULL, NULL, 0, NULL}
@@ -942,10 +1043,10 @@ PyMODINIT_FUNC initaccessibility(void) {
     PyModule_AddObject(m, "DEFAULT_TIMEOUT", PyFloat_FromDouble(0.0));
 #if PY_MAJOR_VERSION >= 3
     PyModule_AddObject(m, "__author__", PyBytes_FromString("Aaron Jacobs <atheriel@gmail.com>"));
-    PyModule_AddObject(m, "__version__", PyBytes_FromString("0.3.2"));
+    PyModule_AddObject(m, "__version__", PyBytes_FromString("0.4.0"));
 #else
     PyModule_AddObject(m, "__author__", PyString_FromString("Aaron Jacobs <atheriel@gmail.com>"));
-    PyModule_AddObject(m, "__version__", PyString_FromString("0.3.2"));
+    PyModule_AddObject(m, "__version__", PyString_FromString("0.4.0"));
 #endif
 
     InvalidUIElementError = PyErr_NewExceptionWithDoc("accessibility.InvalidUIElementError", InvalidUIElementError_docstring, PyExc_ValueError, NULL);
@@ -1109,6 +1210,10 @@ static void handleAXErrors(const char * attribute_name, AXError error) {
             PyErr_SetString(PyExc_KeyError, formattedMessage("This element does not possess the attribute %s.", attribute_name));
             break;
 
+        case kAXErrorActionUnsupported:
+            PyErr_SetString(PyExc_KeyError, formattedMessage("This element does not support the action %s. Note: the system-wide element does not support ANY actions.", attribute_name));
+            break;
+
         case kAXErrorIllegalArgument:
             PyErr_SetString(PyExc_ValueError, formattedMessage("Invalid argument. This is probably caused by a faulty AccessibleElement."));
             break;
